@@ -12,7 +12,30 @@ import logging
 import os
 from datetime import datetime
 from functools import wraps
-from typing import Callable, Dict, Iterable, Union
+from typing import (
+    TYPE_CHECKING,
+    Callable,
+    Dict,
+    Iterable,
+    Optional,
+    Tuple,
+    Union,
+    cast,
+)
+import getpass
+import warnings
+import subprocess
+import sys
+
+
+if TYPE_CHECKING:
+    # https://github.com/microsoft/pyright/issues/1358
+    from typing import TypedDict
+else:
+    try:
+        from typing import TypedDict
+    except ImportError:
+        from typing_extensions import TypedDict
 
 from requests.sessions import Session
 
@@ -20,20 +43,28 @@ from calcbench.api_query_params import APIQueryParams
 
 
 import requests
+from requests import RequestException
 
 logger = logging.getLogger(__name__)
 
 
-try:
-    import pandas as pd
-except ImportError:
-    "Can't find pandas, won't be able to use the functions that return DataFrames."
-    pass
+class _SESSION_VARIABLES(TypedDict):
+    calcbench_user_name: Optional[str]
+    calcbench_password: Optional[str]
+    api_url_base: str
+    logon_url: str
+    domain: str
+    session: Optional[Session]
+    ssl_verify: bool
+    timeout: int
+    enable_backoff: bool
+    proxies: Dict[str, str]
+    backoff_giveup: Optional[Callable[[RequestException], bool]]
 
 
-_SESSION_STUFF = {
-    "calcbench_user_name": os.environ.get("CALCBENCH_USERNAME"),
-    "calcbench_password": os.environ.get("CALCBENCH_PASSWORD"),
+_SESSION_STUFF: _SESSION_VARIABLES = {
+    "calcbench_user_name": None,
+    "calcbench_password": None,
     "api_url_base": "https://www.calcbench.com/api/{0}",
     "logon_url": "https://www.calcbench.com/account/LogOnAjax",
     "domain": "https://www.calcbench.com/{0}",
@@ -41,24 +72,103 @@ _SESSION_STUFF = {
     "session": None,
     "timeout": 60 * 20,  # twenty minute content request timeout, by default
     "enable_backoff": False,
-    "proxies": None,
+    "proxies": {},
+    "backoff_giveup": None,
 }
 
+KEYRING_SERVICE_NAME = "calcbench_api"
+
+USERNAME_ENVIRONMENT_VARIABLE = "CALCBENCH_USERNAME"
+PASSWORD_ENVIRONMENT_VARIABLE = "CALCBENCH_PASSWORD"
+
+
+def _get_credentials() -> Tuple[str, str]:
+    """
+    Used to store credentials as environment variables.  Now we use keyring
+
+    """
+    session_user_name = _SESSION_STUFF.get("calcbench_user_name")
+    session_password = _SESSION_STUFF.get("calcbench_password")
+    if session_user_name and session_password:
+        return (session_user_name, session_password)
+    keyring_found = False
+    try:
+        import keyring
+
+        keyring_found = True
+        logger.info("found keyring module")
+    except ImportError:
+        logger.info("did not find keyring module")
+    else:
+        try:
+            keychain_credentials = keyring.get_credential(KEYRING_SERVICE_NAME, None)
+        except Exception as e:
+            warnings.warn(f"Exception getting credentials from keyring {e}")
+        else:
+            if keychain_credentials:
+                user_name = cast(str, keychain_credentials.username)
+                password = cast(str, keychain_credentials.password)
+                logger.info(f"found credentials for {user_name} in keyring")
+                _SESSION_STUFF["calcbench_user_name"] = user_name
+                _SESSION_STUFF["calcbench_password"] = password
+                return (user_name, password)
+    environment_user_name = os.environ.get(USERNAME_ENVIRONMENT_VARIABLE)
+    environment_password = os.environ.get(PASSWORD_ENVIRONMENT_VARIABLE)
+    if environment_user_name and environment_password:
+        _SESSION_STUFF["calcbench_user_name"] = environment_user_name
+        _SESSION_STUFF["calcbench_password"] = environment_password
+        logger.info(
+            f"found credentials for {environment_user_name} in environment variables"
+        )
+        return (cast(str, environment_user_name), cast(str, environment_password))
+    else:
+        user_name = input("Calcbench username/email::")
+        password = getpass.getpass("Calcbench password::")
+
+        _SESSION_STUFF["calcbench_user_name"] = user_name
+        _SESSION_STUFF["calcbench_password"] = password
+        store_in_keychain = (
+            input(
+                f"Store credentials in keychain? (yes/no).  {'' if keyring_found else 'Will attempt to install the keyring package.'}"
+            ).lower()[0]
+            == "y"
+        )
+
+        if store_in_keychain:
+            if not keyring_found:
+                try:
+                    subprocess.check_call(
+                        [sys.executable, "-m", "pip", "install", "keyring"]
+                    )
+                except Exception as e:
+                    logger.error("exception installing keyring module")
+                else:
+                    logger.info("successfully installed keyring package")
+                    keyring_found = True
+            if keyring_found:
+                import keyring
+
+                try:
+                    keyring.set_password(
+                        KEYRING_SERVICE_NAME, username=user_name, password=password
+                    )
+
+                except Exception as e:
+                    warnings.warn(f"Exception saving credentials to keyring {e}")
+                else:
+                    logger.info("saved credentials to keyring")
+        else:
+            print(
+                f'To avoid this prompt: save credentials to keychain, call "cb.set_credentials()" or set the "{USERNAME_ENVIRONMENT_VARIABLE}" and "{PASSWORD_ENVIRONMENT_VARIABLE}" environment variable'
+            )
+        return (user_name, password)
+
 
 def _calcbench_session() -> Session:
     session = _SESSION_STUFF.get("session")
     if not session:
-        user_name = _SESSION_STUFF.get("calcbench_user_name")
-        password = _SESSION_STUFF.get("calcbench_password")
-        if not (user_name and password):
-            import getpass
+        user_name, password = _get_credentials()
 
-            user_name = input(
-                'Calcbench username/email. Set the "calcbench_user_name" environment variable or call "set_credentials" to avoid this prompt::'
-            )
-            password = getpass.getpass(
-                'Calcbench password.  Set the "calcbench_password" enviroment variable to avoid this prompt::'
-            )
         session = requests.Session()
         if _SESSION_STUFF.get("proxies"):
             session.proxies.update(_SESSION_STUFF["proxies"])
@@ -70,6 +180,8 @@ def _calcbench_session() -> Session:
         )
         r.raise_for_status()
         if r.text != "true":
+            _SESSION_STUFF["calcbench_user_name"] = None
+            _SESSION_STUFF["calcbench_password"] = None
             raise ValueError(
                 "Incorrect Credentials, use the email and password you use to login to Calcbench."
             )
@@ -94,7 +206,8 @@ def _add_backoff(f):
     @wraps(f)
     def wrapper(*args, **kwargs):
         if _SESSION_STUFF["enable_backoff"]:
-            backoff = _SESSION_STUFF["backoff_package"]
+            import backoff
+
             return backoff.on_exception(
                 backoff.expo,
                 requests.exceptions.RequestException,
@@ -110,11 +223,12 @@ def _add_backoff(f):
 
 @_add_backoff
 def _json_POST(end_point: str, payload: Union[dict, APIQueryParams]):
+    session = _calcbench_session()
     url = _SESSION_STUFF["api_url_base"].format(end_point)
     data = json.dumps(payload)
     logger.debug(f"posting to {url}, {data}")
     start = datetime.now()
-    response = _calcbench_session().post(
+    response = session.post(
         url,
         data=data,
         headers={"content-type": "application/json"},
@@ -155,7 +269,7 @@ def set_credentials(cb_username: str, cb_password: str):
 
     Call this before any other Calcbench functions.
 
-    Alternatively set the ``CALCBENCH_USERNAME`` and ``CALCBENCH_PASSWORD`` environment variables
+    Alternatively set the ``CALCBENCH_USERNAME`` and ``CALCBENCH_PASSWORD`` environment variables OR set the calcbench_api credentials on your Keychain|Secret Service|Windows Credential Locker
 
     :param str cb_username: Your calcbench.com email address
     :param str cb_password: Your calcbench.com password
@@ -172,7 +286,8 @@ def set_credentials(cb_username: str, cb_password: str):
 
 def enable_backoff(
     backoff_on: bool = True,
-    giveup: Callable[[Exception], bool] = lambda e: e.response.status_code == 404,
+    giveup: Callable[[RequestException], bool] = lambda e: hasattr(e, "response")
+    and e.response.status_code == 404,
 ):
     """Re-try failed requests with exponential back-off
 
@@ -186,7 +301,7 @@ def enable_backoff(
     :param giveup: function that handles exception and decides whether to continue or not.
 
     Usage::
-        >>> calcbench.enable_backoff(giveup=lambda e: e.response.status_code == 404)
+        >>> calcbench.enable_backoff(giveup=lambda e: e.response and e.response.status_code == 404)
 
     """
     if backoff_on:
@@ -195,11 +310,9 @@ def enable_backoff(
         except ImportError:
             print("backoff package not found, `pip install backoff`")
             raise
-
-        _SESSION_STUFF["backoff_package"] = backoff
+        _SESSION_STUFF["backoff_giveup"] = giveup
 
     _SESSION_STUFF["enable_backoff"] = backoff_on
-    _SESSION_STUFF["backoff_giveup"] = giveup
 
 
 def set_proxies(proxies: Dict[str, str]):
@@ -219,99 +332,6 @@ def set_field_values(dataclass, kwargs: dict, date_columns: Iterable[str] = []):
             setattr(dataclass, k, v)
 
 
-def face_statement(
-    company_identifier,
-    statement_type,
-    period_type="annual",
-    all_history=False,
-    descending_dates=False,
-):
-    """Face Statements.
-
-    face statements as reported by the filing company
-
-
-    :param string company_identifier: a ticker or a CIK code, eg 'msft'
-    :param string statement_type: one of ('income', 'balance', 'cash', 'change-in-equity', 'comprehensive-income')
-    :param string period_type: annual|quarterly|cummulative|combined
-    :param string all_periods: get all history or only the last four, True or False.
-    :param bool descending_dates: return columns in oldest -> newest order.
-
-    :rtype: object
-
-    Returns:
-    An object with columns and line items lists.  The columns have fiscal_period, period_start, period_end and instant values.
-    The line items have label, local_name (the XBRL tag name), tree_depth (the indent amount), is_subtotal (whether or not the line item is computed from above metrics) and facts.
-    The facts are in the same order as the columns and have fact_ids (an internal Calcbench ID), unit_of_measure (USD etc), effective_value (the reported value), and format_type.
-
-
-    Usage::
-        >>> calcbench.face_statement('msft', 'income')
-
-    """
-    url = _SESSION_STUFF["api_url_base"].format("asReported")
-    payload = {
-        "companyIdentifier": company_identifier,
-        "statementType": statement_type,
-        "periodType": period_type,
-        "allPeriods": all_history,
-        "descendingDates": descending_dates,
-    }
-    response = _calcbench_session().get(
-        url,
-        params=payload,
-        headers={"content-type": "application/json"},
-        verify=_SESSION_STUFF["ssl_verify"],
-    )
-    response.raise_for_status()
-    data = response.json()
-    return data
-
-
-as_reported_raw = face_statement
-
-
-def tag_contents(accession_id, block_tag_name):
-    payload = {"accession_ids": accession_id, "block_tag_name": block_tag_name}
-    json = _json_GET("query/disclosuresByTag", payload)
-    return json[0]["blobs"][0]
-
-
-def company_disclosures(ticker, period=None, year=None, statement_type=None):
-    payload = {"companyIdentifier": ticker}
-    if period:
-        payload["period"] = period
-    if year:
-        payload["year"] = year
-    if statement_type:
-        payload["statementType"] = statement_type
-    url = _SESSION_STUFF["api_url_base"].format("companyDisclosures")
-    r = _calcbench_session().get(
-        url, params=payload, verify=_SESSION_STUFF["ssl_verify"]
-    )
-    r.raise_for_status()
-    return r.json()
-
-
-def disclosure_text(network_id):
-    url = _SESSION_STUFF["api_url_base"].format("disclosure")
-    r = _calcbench_session().get(
-        url, params={"networkID": network_id}, verify=_SESSION_STUFF["ssl_verify"]
-    )
-    r.raise_for_status()
-    return r.json()
-
-
-def business_combinations(company_identifiers):
-    payload = {
-        "companiesParameters": {"companyIdentifiers": company_identifiers},
-        "pageParameters": {},
-    }
-    period_parameters = {}
-    payload["periodParameters"] = period_parameters
-    return _json_POST("businessCombinations", payload)
-
-
 def _try_parse_timestamp(timestamp: str):
     """
     We did not always have milliseconds
@@ -325,13 +345,6 @@ def _try_parse_timestamp(timestamp: str):
         return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")
 
 
-def document_types():
-    url = _SESSION_STUFF["api_url_base"].format("documentTypes")
-    r = _calcbench_session().get(url, verify=_SESSION_STUFF["ssl_verify"])
-    r.raise_for_status()
-    return r.json()
-
-
 def html_diff(html_1, html_2):
     """Diff two pieces of html and return a html diff"""
     return _json_POST("textDiff", {"html1": html_1, "html2": html_2})
